""" Module which holds all the important constants used throughout the package.
"""

import logging

from .colors import *
from .gui_const import *
from .parser_const import *
from .paths import *
from .tags import *

logging.getLogger(__name__)
